/*
  map.cpp
  Copyright (C) 2008-2009 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

//#define DEBUG

#include "qgmap.h"
#include "pos.h"
#include "map.h"
#include <qpixmap.h>
#include <qstring.h>
#include <qpixmapcache.h>
#include <stdio.h>
#include <qdatetime.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "map_gmap.h"
#include "map_poi.h"
#include "map_grid.h"

#ifdef QWS
#include <sys/types.h>
#include <unistd.h>
#ifdef DEBUG
#define Dprintf printf("%d:%s ", getpid(), __func__);printf
#else
#define Dprintf if(0) printf
#endif
#else
#include <sys/types.h>
#include <linux/unistd.h>
#include <errno.h>
#ifdef DEBUG
#define Dprintf(...) {printf("%ld:%s ", syscall(__NR_gettid), __func__);printf(__VA_ARGS__);}
#else
#define Dprintf(...)
#endif
#endif

#ifdef THREAD_DEBUG
#define pthread_mutex_init(x, y) {Dprintf("%s(%d) pthread_mutex_init\n", __func__, __LINE__); pthread_mutex_init(x, y);}
#define pthread_mutex_lock(x) {Dprintf("%s(%d) pthread_mutex_lock\n", __func__, __LINE__); pthread_mutex_lock(x);Dprintf("%s(%d) locked\n", __func__, __LINE__);}
#define pthread_mutex_unlock(x) {Dprintf("%s(%d) pthread_mutex_unlock\n", __func__, __LINE__); pthread_mutex_unlock(x);}
#endif

Map::Map(QObject *parent) :
	QObject(parent)
{
	Enter();

	if (g_machine.type() == SL_A300) {
		// 1MB Cache
		// 256x256x32bpp = 256kB/Picture
		// 1MB/256kB = 4 Picture
		cache.setCacheLimit(1024);
	} else {
		// 4MB Cache
		// 256x256x32bpp = 256kB/Picture
		// 4MB/256kB = 16 Picture
		cache.setCacheLimit(4096);
	}

	pthread_mutex_init(&lock, NULL);
	pthread_cond_init(&cond_queue, NULL);
	pthread_cond_init(&cond_process, NULL);

	/* connect signals */
	connect(parent, SIGNAL(configChanged(const QString &, const QString &)),
			this, SIGNAL(configChanged(const QString &, const QString &)));
	connect(parent, SIGNAL(cacheFlush()),
			this, SLOT(flush()));
	connect(this, SIGNAL(updateMap()),
			parent, SIGNAL(updateMap()));

#ifdef SAFE
	result_img.setAutoDelete(TRUE);
#endif

	qt_counter = thread_counter = 0;

	startTimer(100);

	tid = 0;
	thread_run = 0;
}

Map::~Map()
{
	MapItem *item = items.first();
	while (item) {
		delete item;
		item = items.next();
	}

	if (tid) {
		thread_run = 0;

		pthread_mutex_lock(&lock);
		pthread_cond_signal(&cond_queue);
		pthread_mutex_unlock(&lock);

		Dprintf("join...\n");
		pthread_join(tid, NULL);
		Dprintf("joined!\n");
	}
}

QPixmap *Map::getMap(Pos pos)
{
	QString id;
	int max_len = IMAGE_SIZE<<(19-pos.getZoom());
	static QPixmap pix;

	Enter();

	if(tid == 0) {
		if (g_machine.type() == SL_A300) {
			new MapItemGMap(this, 3);
		} else {
			new MapItemGMap(this, 13);
		}
		new MapItemPoi(this);
		new MapItemGrid(this);
		thread_run = 1;
		pthread_create(&tid, 0, &Map::thread_starter, this);
	}

	pos = pos.offset(IMAGE_SIZE/2, IMAGE_SIZE/2);
	pos.normalize();

	if(pos.getY() < 0 || pos.getY() >= max_len){
		pix.resize(256, 256);
		pix.fill(Qt::lightGray);
		return &pix;
	}

	id = pos.serialize();

	pthread_mutex_lock(&lock);

	// find result images
#ifdef SAFE
	QImage *img = result_img.first();
	while (img) {
		pix.convertFromImage(*img);
		Dprintf("caching %s\n", img->text("id").ascii());
		cache.insert(img->text("id"), pix);
		result_img.remove();
		img = result_img.next();
	}
#endif

	// find cache
	if(cache.find(id, pix)){
		Dprintf("in cache %s\n", id.ascii());
		pthread_mutex_unlock(&lock);
		return &pix;
	}

	if (loading.isNull()) {
		loading.load(LOADING_FILE);
	}

	/* queueing Pos */
	if (pos_queue.inQueue(pos)) {
		Dprintf("in queue %s\n", id.ascii());
		pthread_mutex_unlock(&lock);
		return &loading;
	}

	Dprintf("add queue %s\n", id.ascii());
	pos_queue.enqueue(pos);
	pthread_cond_signal(&cond_queue);
	pthread_mutex_unlock(&lock);

	return &loading;
}

void Map::flush()
{
	Enter();
	pthread_mutex_lock(&lock);

	while (is_processing)
		pthread_cond_wait(&cond_process, &lock);

	cache.clear();
#ifdef SAFE
	result_img.clear();
#endif
	pos_queue.clear();
	pthread_mutex_unlock(&lock);
}

void Map::thread()
{
	while (1) {
		Pos pos;
		QImage img;

//		Dprintf("T:lock\n");
		pthread_mutex_lock(&lock);
		while (thread_run && pos_queue.count() == 0) {
//			Dprintf("T:wait queue...\n");
			pthread_cond_wait(&cond_queue, &lock);
//			Dprintf("T:wake up!\n");
		};

		if (!thread_run)
			break;

		Dprintf("T:queue.size = %d\n", pos_queue.count());
		pos = pos_queue.first();
		is_processing = true;

		pthread_mutex_unlock(&lock);
//		Dprintf("T:unlocked\n");

		Dprintf("T:Pos = %s\n", pos.serialize().ascii());

		MapItem *item = items.first();
		while (item) {
			item->drawItem(img, pos);
			item = items.next();
		}

		pthread_mutex_lock(&lock);
		pos_queue.remove(pos);
		is_processing = false;
#ifdef SAFE
		img.setText("id", 0, pos.serialize());
		result_img.append(new QImage(img));
#else
		QPixmap scr;
		scr.convertFromImage(img);
		cache.insert(pos.serialize(), scr);
#endif
		thread_counter++;
		pthread_cond_signal(&cond_process);
		pthread_mutex_unlock(&lock);
	}
}

void *Map::thread_starter(void *arg)
{
	Dprintf("thread_start\n");
	setpriority(PRIO_PROCESS, 0, 10);
	if(arg){
		Map *map = (Map *)arg;
		map->thread();
	}
	Dprintf("thread_end\n");
	return NULL;
}

void Map::timerEvent(QTimerEvent *e)
{
	UNUSED(e);

	int tmp_counter;

	pthread_mutex_lock(&lock);
	tmp_counter = thread_counter;
	pthread_mutex_unlock(&lock);

	if (qt_counter != tmp_counter) {
		qt_counter = tmp_counter;
		emit updateMap();
	}
}

void Map::addItem(MapItem *item)
{
	items.append(item);
}

void Map::setPosition(const Pos &pos)
{
	Enter();
	if(center == pos)
		return;

	pthread_mutex_lock(&lock);
	center = pos;
	pos_queue.setCenter(center);
	pthread_mutex_unlock(&lock);
}

void Map::request(QString const &req, QString const &arg)
{
	MapItem *item = items.first();
	while (item) {
		item->request(req, arg);
		item = items.next();
	}
}
