/*
  sqlite.cpp
  Copyright (C) 2008-2009 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

//#define DEBUG

#include <stdio.h>
#include "qgmap.h"
#include "sqlite.h"
#include <sqlite3.h>

SQLite::SQLite()
{
	db = NULL;
}

SQLite::~SQLite()
{
	close();
}

int SQLite::open(const char *filename)
{
	return sqlite3_open(filename, &db);
}

void SQLite::close()
{
	if(!db)
		return;

#if SQLITE_VERSION_NUMBER > 3006000
	sqlite3_stmt *stmt;
	/* finalize all statement */
	while((stmt = sqlite3_next_stmt(db, 0)) != 0)
		sqlite3_finalize(stmt);
#endif

	/* close connection */
	sqlite3_close(db);
	db = NULL;
}

int SQLite::exec(const char *sql)
{
	int ret;
	char *errmsg = NULL;

	ret = sqlite3_exec(db, sql, NULL, NULL, &errmsg);
	if(errmsg){
		printf("SQLite Err:%s\n", errmsg);
		sqlite3_free(errmsg);
	}

	return ret;
}

int SQLite::execOneColumn(const char *sql)
{
	SQLiteStatement *stmt = newQuery(sql);
	int ret = -1;

	if (!stmt)
		return -1;

	if (stmt->step() != SQLITE_ROW)
		return -1;

	ret = stmt->getColumnInt(0);
	delete stmt;

	return ret;
}

QString SQLite::escape(QString str)
{
	int i;
	QString result;
	while ((i = str.find('\'')) != -1) {
		result += str.left(i) + "''";
		str = str.mid(i + 1);
	}
	result += str;
	return result;
}

SQLiteStatement *SQLite::newQuery(const char *sql)
{
	SQLiteStatement *SQLstmt = new SQLiteStatement();
	int ret;

	ret = sqlite3_prepare(db, sql, -1, &SQLstmt->stmt, NULL);

	if(ret != SQLITE_OK){
		printf("SQLite Err:%s\n", sqlite3_errmsg(db));
		delete SQLstmt;
		return NULL;
	}

	sqlite3_reset(SQLstmt->stmt);

	return SQLstmt;
}

SQLiteStatement::SQLiteStatement()
{
	stmt = NULL;
}
SQLiteStatement::~SQLiteStatement()
{
	if(!stmt)
		return;

	sqlite3_finalize(stmt);
	stmt = NULL;
}

int SQLiteStatement::bind(int index, const unsigned char *data, int size)
{
	return sqlite3_bind_blob(stmt, index, data, size, SQLITE_TRANSIENT);
}

int SQLiteStatement::bind(int index, const QByteArray &data)
{
	return sqlite3_bind_blob(stmt, index, data.data(), data.size(), SQLITE_TRANSIENT);
}

int SQLiteStatement::bind(int index, const char *string)
{
	return sqlite3_bind_text(stmt, index, string, -1, SQLITE_TRANSIENT);
}

int SQLiteStatement::bind(int index, const QString &string)
{
	return sqlite3_bind_text(stmt, index, string.utf8(), -1, SQLITE_TRANSIENT);
}

int SQLiteStatement::bind(int index, double value)
{
	return sqlite3_bind_double(stmt, index, value);
}

int SQLiteStatement::bind(int index, sqlite3_int64 value)
{
	return sqlite3_bind_int64(stmt, index, value);
}

int SQLiteStatement::bind(int index, int value)
{
	return sqlite3_bind_int(stmt, index, value);
}

int SQLiteStatement::bindNull(int index)
{
	return sqlite3_bind_null(stmt, index);
}

int SQLiteStatement::step()
{
	int ret = sqlite3_step(stmt);
	return ret;
}

int SQLiteStatement::reset()
{
	return sqlite3_reset(stmt);
}

QByteArray SQLiteStatement::getColumnBlob(int col)
{
	QByteArray data;

	data.duplicate((const char *)sqlite3_column_blob(stmt, col),
					sqlite3_column_bytes(stmt, col));

	return data;
}

int SQLiteStatement::getColumnInt(int col)
{
	return sqlite3_column_int(stmt, col);
}

double SQLiteStatement::getColumnDouble(int col)
{
	return sqlite3_column_double(stmt, col);
}

QString SQLiteStatement::getColumnText(int col)
{
	return QString::fromUtf8((const char *)sqlite3_column_text(stmt, col));
}

bool SQLiteStatement::isColumnNull(int col)
{
	return sqlite3_column_type(stmt, col) == SQLITE_NULL;
}

void SQLiteStatement::printRow()
{
	int col;
	for (col = 0; col < sqlite3_column_count(stmt); col++) {
		switch (sqlite3_column_type(stmt, col)) {
		case SQLITE_INTEGER:
			printf("%d\t", sqlite3_column_int(stmt, col));
			break;
		case SQLITE_FLOAT:
			printf("%f\t", sqlite3_column_double(stmt, col));
			break;
		case SQLITE_TEXT:
			printf("%s\t", sqlite3_column_text(stmt, col));
			break;
		case SQLITE_BLOB:
			printf("<BLOB>\t");
			break;
		}
	}
	printf("\n");
}

int SQLiteStatement::getRowCount()
{
	printf("sqlite3_data_count(stmt) = %d\n", sqlite3_data_count(stmt));
	printf("sqlite3_column_count(stmt) = %d\n", sqlite3_column_count(stmt));
	return sqlite3_data_count(stmt);
}
